;;;;
;;;; dm-elisp: Difference-based module for emacs-lisp
;;;;					y-ichisugi@aist.go.jp
;;;;					2002-01-29
;;;;
;;;;	Version 0.1 2002-01-29
;;;;		Tested on GNU Emacs 20.7.2

(require 'cl)

(defun dm-inherit-module (from-module to-module)
  "Copy each name in FROM-MODULE into TO-MODULE
unless name collision occurs."
  (maphash (lambda (key val)
	     (let ((fqn (gethash key to-module)))
	       (cond ((null fqn)
		      ;; It's new name. Just inherit it.
		      (setf (gethash key to-module) val))
		     ((symbolp fqn)
		      (if (eq val fqn)
			  ;; Both symbol has the same fqn.
			  ;; That means the name has been already inherited
			  ;; from another super-module.
			  ;; Therefore, do nothing.
			  nil
			  ;; The name is already in the current module.
			  ;; Now, the name becomes ambiguous name.
			  (setf (gethash key to-module)
				(union (if (consp val) val (list val))
				       (list fqn)))))
		     ((consp fqn)
		      ;; The name is already ambiguous name.
		      (union (if (consp val) val (list val))
			     fqn))
		     (t (error "fatal")))
	       ))
	   from-module))


;;; In order to include the name of "dm-require" and "dm-intern" 
;;; to the name-space of top-module, I use the these symbols here.
;;; ( `(dm-require dm-intern) will be removed by the byte-code compiler.) 
(defvar dm-require 'dummy)
(defvar dm-intern 'dummy)

(defvar dm-top-module (make-hash-table)
  "The top-module is a super-module of all other modules.
It's name space is a snapshot of obarray when (require 'dm-elisp)
 is called.")
(mapatoms (lambda (sym) (setf (gethash sym dm-top-module) sym)))

(defvar dm-bottom-module (make-hash-table)
  "The bottom-module is the module
 which should represents the name space for the end-user of emacs,
however, it does not implemented yet.")
(dm-inherit-module dm-top-module dm-bottom-module)

(defvar dm-module-table (make-hash-table)
  "A hashtable whose keys are module names and
whose values are hashtables which represents modules.")
(setf (gethash 'top dm-module-table) dm-top-module)
(setf (gethash 'bottom dm-module-table) dm-bottom-module)

(defvar dm-current-module-name 'bottom
  "Current module name. ")
(defvar dm-current-module dm-bottom-module
  "Current module.")


(defun dm-require (module)
  "Load the program file, MODULE written in the dm-elisp language.
The MODULE should be a string.
The current module inherits all names visible in the required module.
Currently, the load-path is not examined.
\"./MODULE.el\" is loaded."
  (let* ((module-name (intern module))
	 (super-module (or (gethash module-name dm-module-table)
			  (dm-load-super-module module-name))))
    (dm-inherit-module super-module dm-current-module)
    ))

(defun dm-load-super-module (module-name)
  (let ((super-module (make-hash-table)))
    (setf (gethash module-name dm-module-table) super-module)
    ;; The top-module becomes a super-module by default.
    (dm-inherit-module dm-top-module super-module)
    (let ((dm-current-module-name module-name)
	  (dm-current-module super-module)
	  (file-name (concat (symbol-name module-name) ".el")))
      (if (not (file-exists-p file-name))
	  (error "File %s does not exist." file-name))
      ;; Setup of the new name space has been completed.
      ;; Now, read and eval s-expressions contained in the file.
      (save-excursion
	(find-file (dm-find-library-file file-name))
	(goto-char 0)
	(let ((buf (current-buffer))
	      sexp)
	  (catch 'eof
	    (while t
	      (condition-case ()
		  (setq sexp (read buf))
		(error
		 (kill-buffer (current-buffer))
		 (throw 'eof nil)))
	      ;;(message "read:%S" sexp)	; for debug
	      (setq sexp (dm-trans-sexp sexp))
	      ;;(message "trans:%S" sexp) ; for debug
	      (eval sexp))
	    ))))
    super-module))

(defun dm-find-library-file (file-name)
  file-name)

(defun dm-trans-sexp (sexp)
  (cond
      ((null sexp) sexp)
      ((consp sexp) (cons (dm-trans-sexp (car sexp))
			  (dm-trans-sexp (cdr sexp))))
      ((symbolp sexp) (dm-trans-symbol sexp))
      ((vectorp sexp)
       (dotimes (n (length sexp))
	 (setf (aref sexp n) (dm-trans-sexp (aref sexp n))))
       sexp)
      ((atom sexp) sexp)
      (t (error "Unknown sexp: %s" sexp))))

(defun dm-trans-symbol (sym)
  (if (string-match ":" (symbol-name sym))
      ;; FQN notation "m:n" is returned directly.
      ;; (Current implementation does not check if the prefix is
      ;;   a name of one of super-modules of the current module.)
      sym
      ;; Intern the name to the current module.
      (dm-intern sym)))

(defun dm-intern (simple-name &optional module-name)
  "Intern SIMPLE-NAME to the module specified by MODULE-NAME.
SIMPLE-NAME may be a symbol or a string.
If MODULE-NAME is omitted, the value of dm-current-module is used."
  (if (null module-name)
      (setq module-name dm-current-module-name))
  (if (stringp simple-name)
      (setq simple-name (intern simple-name)))
  (let* ((module (or (gethash module-name dm-module-table)
		     (error "Module %s not found." module-name)))
	 (fqn (gethash simple-name module)))
    (cond ((null fqn)
	   (let ((new-fqn (intern (concat (symbol-name module-name)
					   ":"
					   (symbol-name simple-name)))))
	     (setf (gethash simple-name module) new-fqn)
	     new-fqn))
	  ((symbolp fqn)
	   fqn)
	  ((consp fqn)
	   (error "Ambiguous reference: %s" fqn))
	  (t (error "fatal")))))
  

(defun dm-print-module (&optional module)
  "Print the all symbols in MODULE.
Just for debugging of this module mechanism."
  (if (stringp module)
      (setq module (intern module)))
  (cond ((null module)
	 (setq module dm-current-module))
	((symbolp module)
	 (setq module (or (gethash module dm-module-table)
			  (error "Module %s not found." module)))))
  (maphash (lambda (key val)
	     (print (list key val))
	     )
	   module))

(provide 'dm-elisp)
